/* can_sysdep.h - hides differences between individual Linux kernel
 *                versions and RT extensions
 * Linux CAN-bus device driver.
 * Written by Pavel Pisa - OCERA team member
 * email:pisa@cmp.felk.cvut.cz
 * This software is released under the GPL-License.
 * Version lincan-0.3  17 Jun 2004
 */

#ifndef _CAN_SYSDEP_H
#define _CAN_SYSDEP_H

#ifdef CAN_WITH_RTL
#include <rtl.h>
#include <rtl_sync.h>
#include <rtl_core.h>
#include <rtl_mutex.h>
#include <rtl_sched.h>
#include <time.h>
#endif /*CAN_WITH_RTL*/

#include <cpu_def.h>
#include <malloc.h>

// typedef unsigned long atomic_t;
typedef struct { volatile int counter; } atomic_t;

#define mb()	__memory_barrier()

/**
 * container_of - cast a member of a structure out to the containing structure
 * @ptr:	the pointer to the member.
 * @type:	the type of the container struct this is embedded in.
 * @member:	the name of the member within the struct.
 *
 */
#define container_of(ptr, type, member) ({			\
	const typeof( ((type *)0)->member ) *__mptr = (ptr);	\
	(type *)( (char *)__mptr - offsetof(type,member) );})


/*#define __NO_VERSION__*/
/*#include <linux/module.h>*/

// #include <linux/version.h>
// #include <linux/wait.h>
// #include <linux/list.h>
// #include <linux/fs.h>
// #include <linux/ioport.h>
// #include <linux/delay.h>
// #include <linux/sched.h>
// #include <linux/interrupt.h>
// #include <asm/errno.h>
//
// #include <asm/io.h>
// #include <asm/atomic.h>
// #include <asm/irq.h>
// #include <asm/uaccess.h>

// #include "lincan_config.h"

/*optional features*/
// #if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0))
// #define CAN_ENABLE_KERN_FASYNC
// #ifdef CONFIG_PCI
// #define CAN_ENABLE_PCI_SUPPORT
// #endif
// #ifdef CONFIG_OC_LINCANVME
// #define CAN_ENABLE_VME_SUPPORT
// #endif
// #endif

// #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
// #include <linux/malloc.h>
// #else
// #include <linux/slab.h>
// #endif

// #ifdef CAN_ENABLE_PCI_SUPPORT
// #include "linux/pci.h"
// #endif /*CAN_ENABLE_PCI_SUPPORT*/

/* Next is not sctrictly correct, because of 2.3.0, 2.3.1, 2.3.2
   kernels need next definitions  too */
// #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,2,19)) /* may need correction */
   #define wait_queue_head_t struct wait_queue *
   #define wait_queue_t      struct wait_queue
   #define init_waitqueue_head(queue_head) (*queue_head=NULL)
  #define init_waitqueue_entry(qentry,qtask) \
                        (qentry->next=NULL,qentry->task=qtask)
  #define DECLARE_WAIT_QUEUE_HEAD(name) \
        struct wait_queue * name=NULL
  #define DECLARE_WAITQUEUE(wait, current) \
        struct wait_queue wait = { current, NULL }
//   #define init_MUTEX(sem) (*sem=MUTEX)
//   #define DECLARE_MUTEX(name) struct semaphore name=MUTEX
// #endif /* 2.2.19 */

// #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)) && !defined(DECLARE_TASKLET)
//   #define tasklet_struct tq_struct
/*  #define DECLARE_TASKLET(_name, _func, _data) \
                struct tq_struct _name = { sync: 0, routine: _func, data: (void*)_data }*/
//
//   /* void tasklet_init(struct tasklet_struct *t, void (*func)(unsigned long), unsigned long data); */
/*   #define tasklet_init(_tasklet, _func, _data) \
    do{ \
       (_tasklet)->sync=0; \
       (_tasklet)->routine=_func; \
       (_tasklet)->data=(void*)_data; \
    }while(0)*/
//
//   /* void tasklet_schedule(struct tasklet_struct *t) */
/*  #define tasklet_schedule(_tasklet) \
    do{ \
       queue_task(_tasklet,&tq_immediate); \
       mark_bh(IMMEDIATE_BH); \
    }while(0)*/
//
//   /* void tasklet_kill(struct tasklet_struct *t); */
/*  #define tasklet_kill(_tasklet) \
                synchronize_irq()*/
//
// #endif /* 2.4.0 */


// #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,7)) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))

#define MINOR_NR \
	(MINOR(file->f_dentry->d_inode->i_rdev))

// #else /* Linux kernel < 2.5.7 or >= 2.6.0 */
//
/*#define MINOR_NR \
	(minor(file->f_dentry->d_inode->i_rdev))*/
//
// #endif /* Linux kernel < 2.5.7 or >= 2.6.0 */

// #ifndef CAN_WITH_RTL
// #if ((LINUX_VERSION_CODE < KERNEL_VERSION(2,5,68)) && !defined(IRQ_RETVAL))
  typedef void can_irqreturn_t;
  #define CAN_IRQ_NONE
  #define CAN_IRQ_HANDLED
  #define CAN_IRQ_RETVAL(x)
// #else /* <=2.5.67 */
//   typedef irqreturn_t can_irqreturn_t;
//   #define CAN_IRQ_NONE    IRQ_NONE
//   #define CAN_IRQ_HANDLED IRQ_HANDLED
//   #define CAN_IRQ_RETVAL  IRQ_RETVAL
// #endif /* <=2.5.67 */
// #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
/*  #define CAN_IRQ_HANDLER_ARGS(irq_number, dev_id) \
		int irq_number, void *dev_id, struct pt_regs *regs*/
// #else /* < 2.6.19 */
  #define CAN_IRQ_HANDLER_ARGS(irq_number, dev_id) \
		int irq_number, void *dev_id
// #endif /* < 2.6.19 */
// #else /*CAN_WITH_RTL*/
//   typedef int can_irqreturn_t;
//   #define CAN_IRQ_NONE        0
//   #define CAN_IRQ_HANDLED     1
//   #define CAN_IRQ_RETVAL(x)   ((x) != 0)
/*  #define CAN_IRQ_HANDLER_ARGS(irq_number, dev_id) \
		int irq_number, void *dev_id, struct pt_regs *regs*/
// #endif /*CAN_WITH_RTL*/

// #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,33))
//    #define can_synchronize_irq(irqnum) synchronize_irq()
// #else /* >=2.5.33 */
   #define can_synchronize_irq synchronize_irq
// #endif

// #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
//   #define del_timer_sync del_timer
// #endif /* <2.4.0 */

// #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9))
   typedef unsigned long can_ioptr_t;
   #define can_ioptr2ulong(ioaddr) ((unsigned long)(ioaddr))
   #define can_ulong2ioptr(addr)   ((unsigned long)(addr))
   #define can_inb(ioaddr) inb(ioaddr)
   #define can_outb(data,ioaddr) outb(data,ioaddr)
   #define can_inw(ioaddr) inb(ioaddr)
   #define can_outw(data,ioaddr) outb(data,ioaddr)
   #define can_inl(ioaddr) inb(ioaddr)
   #define can_outl(data,ioaddr) outb(data,ioaddr)
// #else /* >=2.6.9 */
//    typedef void __iomem * can_ioptr_t;
//    #define can_ioptr2ulong(ioaddr) ((unsigned long __force)(ioaddr))
//    #define can_ulong2ioptr(addr)   ((can_ioptr_t)(addr))
//    #define can_inb(ioaddr) inb(can_ioptr2ulong(ioaddr))
//    #define can_outb(data,ioaddr) outb(data,can_ioptr2ulong(ioaddr))
//    #define can_inw(ioaddr) inb(can_ioptr2ulong(ioaddr))
//    #define can_outw(data,ioaddr) outb(data,can_ioptr2ulong(ioaddr))
//    #define can_inl(ioaddr) inb(can_ioptr2ulong(ioaddr))
//    #define can_outl(data,ioaddr) outb(data,can_ioptr2ulong(ioaddr))
// #endif

#define can_readb  readb
#define can_writeb writeb
#define can_readw  readw
#define can_writew writew
#define can_readl  readl
#define can_writel writel

#define can_ioport2ioptr can_ulong2ioptr

#ifdef __HAVE_ARCH_CMPXCHG
  #define CAN_HAVE_ARCH_CMPXCHG
#endif

// #ifndef CAN_WITH_RTL
/* Standard LINUX kernel */

#define can_spinlock_t             long
#define can_spin_irqflags_t        unsigned long
#define can_spin_lock(lock)        cli()
#define can_spin_unlock(lock)      sti()
#define can_spin_lock_irqsave(lock,flags)      save_and_cli(flags)
#define can_spin_unlock_irqrestore(lock,flags) restore_flags(flags)
#define can_spin_lock_init         can_splck_init

#define CAN_DEFINE_SPINLOCK(x)     can_spinlock_t x = 0

static inline
void can_splck_init(can_spinlock_t *x)
{
	*x=0;
}

// #if !defined(CONFIG_PREEMPT_RT) && ( defined(CONFIG_PREEMPT) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)) )
// #define can_preempt_disable        preempt_disable
// #define can_preempt_enable         preempt_enable
// #else /*CONFIG_PREEMPT*/
#define can_preempt_disable()      do { } while (0)
#define can_preempt_enable()       do { } while (0)
// #endif /*CONFIG_PREEMPT*/

// #define can_enable_irq             sti()
// #define can_disable_irq            cli()
#define can_enable_irq(var)           (var=1)
#define can_disable_irq(var)          (var=0)

#define can_printk                 printf
#define KERN_CRIT
#define KERN_ERR

/// LINUX src: include/asm-arm/bitops.h

#define set_bit ____atomic_set_bit
#define clear_bit ____atomic_clear_bit
#define change_bit ____atomic_change_bit
#define test_and_set_bit ____atomic_test_and_set_bit
#define test_and_clear_bit ____atomic_test_and_clear_bit
#define test_and_change_bit ____atomic_test_and_change_bit
#define raw_local_irq_save(flags) save_and_cli(flags);
#define raw_local_irq_restore(flags) restore_flags(flags);
/*
 * These functions are the basis of our bit ops.
 *
 * First, the atomic bitops. These use native endian.
 */
static inline void ____atomic_set_bit(unsigned int bit, volatile unsigned long *p)
{
	unsigned long flags;
	unsigned long mask = 1UL << (bit & 31);

	p += bit >> 5;

	raw_local_irq_save(flags);
	*p |= mask;
	raw_local_irq_restore(flags);
}

static inline void ____atomic_clear_bit(unsigned int bit, volatile unsigned long *p)
{
	unsigned long flags;
	unsigned long mask = 1UL << (bit & 31);

	p += bit >> 5;

	raw_local_irq_save(flags);
	*p &= ~mask;
	raw_local_irq_restore(flags);
}

static inline void ____atomic_change_bit(unsigned int bit, volatile unsigned long *p)
{
	unsigned long flags;
	unsigned long mask = 1UL << (bit & 31);

	p += bit >> 5;

	raw_local_irq_save(flags);
	*p ^= mask;
	raw_local_irq_restore(flags);
}

static inline int
____atomic_test_and_set_bit(unsigned int bit, volatile unsigned long *p)
{
	unsigned long flags;
	unsigned int res;
	unsigned long mask = 1UL << (bit & 31);

	p += bit >> 5;

	raw_local_irq_save(flags);
	res = *p;
	*p = res | mask;
	raw_local_irq_restore(flags);

	return res & mask;
}

static inline int
____atomic_test_and_clear_bit(unsigned int bit, volatile unsigned long *p)
{
	unsigned long flags;
	unsigned int res;
	unsigned long mask = 1UL << (bit & 31);

	p += bit >> 5;

	raw_local_irq_save(flags);
	res = *p;
	*p = res & ~mask;
	raw_local_irq_restore(flags);

	return res & mask;
}

static inline int
____atomic_test_and_change_bit(unsigned int bit, volatile unsigned long *p)
{
	unsigned long flags;
	unsigned int res;
	unsigned long mask = 1UL << (bit & 31);

	p += bit >> 5;

	raw_local_irq_save(flags);
	res = *p;
	*p = res ^ mask;
	raw_local_irq_restore(flags);

	return res & mask;
}

/// LINUX src: include/asm-arm/atomic.h

#define atomic_read(v)	((v)->counter)

#define atomic_set(v,i)	(((v)->counter) = (i))

static inline int atomic_add_return(int i, atomic_t *v)
{
	unsigned long flags;
	int val;

	raw_local_irq_save(flags);
	val = v->counter;
	v->counter = val += i;
	raw_local_irq_restore(flags);

	return val;
}

static inline int atomic_sub_return(int i, atomic_t *v)
{
	unsigned long flags;
	int val;

	raw_local_irq_save(flags);
	val = v->counter;
	v->counter = val -= i;
	raw_local_irq_restore(flags);

	return val;
}

static inline int atomic_cmpxchg(atomic_t *v, int old, int new)
{
	int ret;
	unsigned long flags;

	raw_local_irq_save(flags);
	ret = v->counter;
	if (ret == old)
		v->counter = new;
	raw_local_irq_restore(flags);

	return ret;
}

#define atomic_xchg(v, new) (xchg(&((v)->counter), new))

static inline int atomic_add_unless(atomic_t *v, int a, int u)
{
	int c, old;

	c = atomic_read(v);
	while (c != u && (old = atomic_cmpxchg((v), c, c + a)) != c)
		c = old;
	return c != u;
}
#define atomic_inc_not_zero(v) atomic_add_unless((v), 1, 0)

#define atomic_add(i, v)	(void) atomic_add_return(i, v)
#define atomic_inc(v)		(void) atomic_add_return(1, v)
#define atomic_sub(i, v)	(void) atomic_sub_return(i, v)
#define atomic_dec(v)		(void) atomic_sub_return(1, v)

#define atomic_inc_and_test(v)	(atomic_add_return(1, v) == 0)
#define atomic_dec_and_test(v)	(atomic_sub_return(1, v) == 0)
#define atomic_inc_return(v)    (atomic_add_return(1, v))
#define atomic_dec_return(v)    (atomic_sub_return(1, v))
#define atomic_sub_and_test(i, v) (atomic_sub_return(i, v) == 0)

#define atomic_add_negative(i,v) (atomic_add_return(i, v) < 0)

/* Atomic operations are already serializing on ARM */
#define smp_mb__before_atomic_dec()	barrier()
#define smp_mb__after_atomic_dec()	barrier()
#define smp_mb__before_atomic_inc()	barrier()
#define smp_mb__after_atomic_inc()	barrier()

/// LINUX src: include/linux/interrupt.h

#define ATOMIC_INIT(i)	{ (i) }

/* Tasklets --- multithreaded analogue of BHs.

   Main feature differing them of generic softirqs: tasklet
   is running only on one CPU simultaneously.

   Main feature differing them of BHs: different tasklets
   may be run simultaneously on different CPUs.

   Properties:
   * If tasklet_schedule() is called, then tasklet is guaranteed
     to be executed on some cpu at least once after this.
   * If the tasklet is already scheduled, but its excecution is still not
     started, it will be executed only once.
   * If this tasklet is already running on another CPU (or schedule is called
     from tasklet itself), it is rescheduled for later.
   * Tasklet is strictly serialized wrt itself, but not
     wrt another tasklets. If client needs some intertask synchronization,
     he makes it with spinlocks.
 */

struct tasklet_struct
{
	struct tasklet_struct *next;
	unsigned long state;
	atomic_t count;
	void (*func)(unsigned long);
	unsigned long data;
};

#define DECLARE_TASKLET(name, func, data) \
struct tasklet_struct name = { NULL, 0, ATOMIC_INIT(0), func, data }

#define DECLARE_TASKLET_DISABLED(name, func, data) \
struct tasklet_struct name = { NULL, 0, ATOMIC_INIT(1), func, data }

/* CAN message timestamp source, it is called from interrupt context */
//#define can_gettimeofday do_gettimeofday

/// from linux/timer.h

struct tvec_t_base_s;

struct timer_list {
	struct list_head entry;
	unsigned long expires;

	void (*function)(unsigned long);
	unsigned long data;

	struct tvec_t_base_s *base;
#ifdef CONFIG_TIMER_STATS
	void *start_site;
	char start_comm[16];
	int start_pid;
#endif
};

static inline void udelay(long time)
{
	volatile long ticks=(time * CCLK) / 2000000;
	do{
		ticks--;
	}
	while(ticks>0);
}

// #else /*CAN_WITH_RTL*/
//
// #define can_spinlock_t             long
// #define can_spin_irqflags_t        unsigned long
// #define can_spin_lock              save_and_cli
// #define can_spin_unlock            restore_flags
// #define can_spin_lock_irqsave      save_and_cli
// #define can_spin_unlock_irqrestore restore_flags
// #define can_spin_lock_init         can_splck_init
//
// #define CAN_DEFINE_SPINLOCK(x)     can_spinlock_t x = 0
//
// #define can_preempt_disable()      do { } while (0)
// #define can_preempt_enable()       do { } while (0)
//
// #define can_enable_irq             sti
// #define can_disable_irq            cli
//
// #define can_printk                 rtl_printf
//
// /*
//  * terrible hack to test rtl_file private_data concept, ugh !!!
//  * this would result in crash on architectures,  where
//  * sizeof(int) < sizeof(void *)
//  */
// #define can_set_rtl_file_private_data(fptr, p) do{ fptr->f_minor=(long)(p); } while(0)
// #define can_get_rtl_file_private_data(fptr) ((void*)((fptr)->f_minor))
//
// extern can_spinlock_t can_irq_manipulation_lock;
//
// /* CAN message timestamp source, it is called from interrupt context */
/*#define can_gettimeofday(ptr) do {\
	  struct timespec temp_timespec;\
	  clock_gettime(CLOCK_REALTIME,&temp_timespec);\
	  ptr->tv_usec=temp_timespec.tv_nsec/1000;\
	  ptr->tv_sec=temp_timespec.tv_sec;\
	} while(0)*/
//
// #endif /*CAN_WITH_RTL*/

#endif /*_CAN_SYSDEP_H*/
